import pytest

from unittest.mock import call, patch

from redis_key_tagging.client import RedisKeyTagging


class TestRedisKeyTagging:
    @pytest.mark.parametrize(
        "tag, expected_tag_key",
        (
            ("foo", "tag:foo"),
            ("bar", "tag:bar"),
            ("123", "tag:123"),
            ("foo bar", None),
            ("", None),
            ([], None),
            (None, None),
            (True, None),
            (False, None),
        ),
    )
    def test_get_tag_key(self, tag, expected_tag_key):
        redis = RedisKeyTagging()
        assert redis.get_tag_key(tag) == expected_tag_key

    @pytest.mark.parametrize(
        "key, tags, expected_calls",
        (
            ("foo", [], []),
            ("foo", [None], []),
            ("foo", ["bar"], [call().sadd("tag:bar", "foo")]),
            (
                "foo",
                ["bar", "baz"],
                [call().sadd("tag:bar", "foo"), call().sadd("tag:baz", "foo")],
            ),
        ),
    )
    @patch("redis_key_tagging.client.RedisKeyTagging.pipeline")
    @patch("redis.client.Redis.set")
    def test_set(self, mock_set, mock_pipeline, key, tags, expected_calls):
        mock_set.return_value = 1
        redis = RedisKeyTagging()
        result = redis.set(key, "", tags=tags)
        assert result == 1
        redis.pipeline.assert_has_calls([call()])
        redis.pipeline.assert_has_calls(expected_calls, any_order=True)
        redis.pipeline.assert_has_calls([call().execute()])

    @pytest.mark.parametrize(
        "tag, smembers, expected_calls",
        (
            ("foo", {}, []),
            ("foo", {b"bar"}, [call().delete("bar"), call().srem("tag:foo", "bar")]),
            (
                "foo",
                {b"bar", b"baz"},
                [call().delete("bar", "baz"), call().srem("tag:foo", "bar", "baz")],
            ),
        ),
    )
    @patch("redis_key_tagging.client.RedisKeyTagging.smembers")
    @patch("redis_key_tagging.client.RedisKeyTagging.pipeline")
    def test_delete_keys_by_tag(self, mock_pipeline, mock_smembers, tag, smembers, expected_calls):
        mock_pipeline.return_value.execute.return_value = [len(expected_calls)] * 2
        mock_smembers.return_value = sorted(smembers)
        redis = RedisKeyTagging()
        result = redis.delete_keys_by_tag(tag)
        assert result == [len(expected_calls)] * 2
        redis.pipeline.assert_has_calls([call()])
        redis.pipeline.assert_has_calls(expected_calls, any_order=True)
        redis.pipeline.assert_has_calls([call().execute()])



from redis import Redis


class RedisKeyTagging(Redis):
    """
    A Redis client with key tagging and invalidation by tag features.
    """

    TAG_KEY_PREFIX = "tag:"

    def get_tag_key(self, tag):
        """
        Get the key for a given ``tag`` by prexing it with the ``TAG_KEY_PREFIX`` class attribute.
        """
        if isinstance(tag, str) and tag.isalnum():
            return f"{self.TAG_KEY_PREFIX}{tag}"
        else:
            return None

    def set(self, name, value, ex=None, px=None, nx=False, xx=False, tags=[]):
        """
        Set the value at key ``name`` to ``value``

        ``ex`` sets an expire flag on key ``name`` for ``ex`` seconds.

        ``px`` sets an expire flag on key ``name`` for ``px`` milliseconds.

        ``nx`` if set to True, set the value at key ``name`` to ``value`` only
            if it does not exist.

        ``xx`` if set to True, set the value at key ``name`` to ``value`` only
            if it already exists.

        ``tags`` add the key ``name`` to every tag in the ``tags`` list.
        """
        pipeline = self.pipeline()
        for tag in tags:
            tag_key = self.get_tag_key(tag)
            if tag_key:
                pipeline.sadd(tag_key, name)
        pipeline.execute()
        return super().set(name, value, ex, px, nx, xx)

    def delete_keys_by_tag(self, tag):
        """
        Delete all keys from a given tag and remove those keys from the tag.
        """
        pipeline = self.pipeline()
        tag_key = self.get_tag_key(tag)
        if tag_key:
            keys = [key.decode() for key in self.smembers(tag_key)]
            if keys:
                pipeline.delete(*keys)
                pipeline.srem(tag_key, *keys)
        return pipeline.execute()
